#!/usr/bin/perl
use strict;
use warnings;
use Test::More;

use Test::Spelling;
use Pod::Wordlist;

add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib ) );

__DATA__
linter
perltidy
reformats
SUBCOMMANDS
yamllint
yamltidy
Müller
